# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# flake8: noqa

"""Utilities for generating text."""

import json
import sys
from collections.abc import Iterable
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F

if TYPE_CHECKING:
    import nemo.collections.multimodal.speech_llm.modules.common.text_generation_utils as text_generation_utils

from nemo.collections.common.tokenizers.tabular_tokenizer import TabularTokenizer
from nemo.collections.multimodal.speech_llm.modules.common.audio_text_generation_strategy import (
    model_inference_strategy_dispatcher,
)
from nemo.utils import AppState, logging

try:
    from megatron.core import parallel_state, tensor_parallel

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):

    HAVE_MEGATRON_CORE = False

try:
    from megatron.core.num_microbatches_calculator import reconfigure_num_microbatches_calculator

except (ImportError, ModuleNotFoundError):
    logging.warning("Megatron num_microbatches_calculator not found, using Apex version.")
    from apex.transformer.pipeline_parallel.utils import (
        _reconfigure_microbatch_calculator as reconfigure_num_microbatches_calculator,
    )

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

__all__ = [
    "get_computeprob_response",
    "generate",
]


class OutputType(TypedDict):
    sentences: List[str]  # output sentences
    tokens: List[List[str]]  # output sentences borken into tokens
    logprob: List[List[float]]  # log prob of generated tokens
    full_logprob: List[List[float]]  # log prob of all the tokens in the vocab
    token_ids: List[List[int]]  # output sentence token ids
    offsets: List[List[int]]  # list of tokens start positions in text


def get_computeprob_response(tokenizer, response, inputs):
    return text_generation_utils.get_computeprob_response(tokenizer, response, inputs)


def send_generate_info(
    context_tokens_tensor,
    context_length_tensor,
    audio_signal,
    audio_signal_length,
    tokens_to_generate,
    all_probs,
    compute_logprob,
    temperature,
    top_k,
    top_p,
    greedy,
    repetition_penalty,
    min_tokens_to_generate,
    end_strings,
    num_audios: Optional[torch.Tensor] = None,
    context_start_idx: Optional[List[List[int]]] = None,
):
    """
    Needs to be synced up with receive_generate_info
    """
    model_parallel_group = parallel_state.get_model_parallel_group()
    src = text_generation_utils.get_model_parallel_src_rank()

    audio_max_len = audio_signal.size(1) if audio_signal is not None else 0

    # Send the sizes of the tensors
    input_info = [
        context_tokens_tensor.size(0),  # batch_size
        context_tokens_tensor.size(1),  # seq_len
        audio_max_len,  # audio_max_len
        tokens_to_generate,
        all_probs,
        compute_logprob,  # whether to compute log probabilities matrix
        temperature,
        top_k,
        top_p,
        greedy,
        repetition_penalty,
        min_tokens_to_generate,
    ]
    input_info_tensor = torch.cuda.FloatTensor(input_info)
    torch.distributed.broadcast(input_info_tensor, src, model_parallel_group)

    # Send variables to all ranks
    torch.distributed.broadcast(context_length_tensor, src, model_parallel_group)
    torch.distributed.broadcast(context_tokens_tensor, src, model_parallel_group)

    torch.distributed.broadcast(audio_signal, src, model_parallel_group)
    torch.distributed.broadcast(audio_signal_length, src, model_parallel_group)

    # send end strings
    string_tensor = torch.as_tensor(
        np.frombuffer(json.dumps(end_strings).encode('utf-8'), dtype=np.int8), device=torch.cuda.current_device()
    )
    size = torch.as_tensor([string_tensor.size(0)], device=torch.cuda.current_device(), dtype=torch.int64)
    torch.distributed.broadcast(size, src, model_parallel_group)
    torch.distributed.broadcast(string_tensor, src, model_parallel_group)

    if num_audios is not None:
        torch.distributed.broadcast(num_audios, src, model_parallel_group)

    if context_start_idx is not None:
        context_idx_tensor = torch.as_tensor(
            np.frombuffer(json.dumps(context_start_idx).encode('utf-8'), dtype=np.int8),
            device=torch.cuda.current_device(),
        )
        ctx_size = torch.as_tensor([context_idx_tensor.size(0)], device=torch.cuda.current_device(), dtype=torch.int64)
        torch.distributed.broadcast(ctx_size, src, model_parallel_group)
        torch.distributed.broadcast(context_idx_tensor, src, model_parallel_group)


def receive_generate_info(has_multi_audios=False):
    """
    Needs to be synced up with send_generate_info
    """
    model_parallel_group = parallel_state.get_model_parallel_group()
    src = text_generation_utils.get_model_parallel_src_rank()
    input_info_tensor = torch.empty(12, dtype=torch.float32, device=torch.cuda.current_device())
    torch.distributed.broadcast(input_info_tensor, src, model_parallel_group)
    batch_size = int(input_info_tensor[0].item())
    seq_len = int(input_info_tensor[1].item())
    audio_len = int(input_info_tensor[2].item())
    tokens_to_generate = int(input_info_tensor[3].item())
    all_probs = bool(input_info_tensor[4].item())
    compute_logprob = bool(input_info_tensor[5].item())  # whether to compute log probabilities matrix
    temperature = float(input_info_tensor[6].item())
    top_k = int(input_info_tensor[7].item())
    top_p = float(input_info_tensor[8].item())
    greedy = bool(input_info_tensor[9].item())
    repetition_penalty = float(input_info_tensor[10].item())
    min_tokens_to_generate = int(input_info_tensor[11].item())

    context_length_tensor = torch.empty(batch_size, dtype=torch.int64, device=torch.cuda.current_device())
    context_tokens_tensor = torch.empty(batch_size, seq_len, dtype=torch.int64, device=torch.cuda.current_device())
    # Send variables to all ranks
    torch.distributed.broadcast(context_length_tensor, src, model_parallel_group)
    torch.distributed.broadcast(context_tokens_tensor, src, model_parallel_group)

    audio_signal = torch.empty(batch_size, audio_len, dtype=torch.float32, device=torch.cuda.current_device())
    audio_signal_length = torch.empty(batch_size, dtype=torch.int64, device=torch.cuda.current_device())
    # Send variables to all ranks
    torch.distributed.broadcast(audio_signal, src, model_parallel_group)
    torch.distributed.broadcast(audio_signal_length, src, model_parallel_group)

    array_size = torch.empty(1, dtype=torch.int64, device=torch.cuda.current_device())
    torch.distributed.broadcast(array_size, src, model_parallel_group)

    string_tensor = torch.empty(array_size[0], dtype=torch.int8, device=torch.cuda.current_device())
    torch.distributed.broadcast(string_tensor, src, model_parallel_group)
    bytes = string_tensor.cpu().numpy().tobytes()
    end_strings = json.loads(bytes.decode('utf-8'))

    num_audios = None
    context_start_idx = None
    if has_multi_audios:
        num_audios = torch.empty(batch_size, dtype=torch.int64, device=torch.cuda.current_device())
        torch.distributed.broadcast(num_audios, src, model_parallel_group)

        array_size = torch.empty(1, dtype=torch.int64, device=torch.cuda.current_device())
        torch.distributed.broadcast(array_size, src, model_parallel_group)
        context_idx_tensor = torch.empty(array_size[0], dtype=torch.int8, device=torch.cuda.current_device())
        torch.distributed.broadcast(context_idx_tensor, src, model_parallel_group)
        bytes = context_idx_tensor.cpu().numpy().tobytes()
        context_start_idx = json.loads(bytes.decode('utf-8'))

    return (
        context_length_tensor,
        context_tokens_tensor,
        audio_signal,
        audio_signal_length,
        tokens_to_generate,
        all_probs,
        compute_logprob,
        temperature,
        top_k,
        top_p,
        greedy,
        repetition_penalty,
        min_tokens_to_generate,
        end_strings,
        num_audios,
        context_start_idx,
    )


def synced_generate(
    model,
    inference_strategy,
    context_tokens_tensor,
    context_length_tensor,
    audio_signal,
    audio_signal_length,
    tokens_to_generate,
    all_probs,
    temperature,
    top_k=0,
    top_p=0.0,
    greedy=False,
    compute_attention_mask=True,
    compute_logprob=False,
    repetition_penalty=1.2,
    end_strings=[],
    min_tokens_to_generate=0,
    num_audios: Optional[torch.Tensor] = None,
    context_start_idx: Optional[List[List[int]]] = None,
):
    context_length = context_length_tensor.min().item()
    tokenizer = model.tokenizer
    if isinstance(tokenizer, TabularTokenizer):
        raise NotImplementedError("Tabular generation is not supported yet")
    else:
        batch_token_iterator = sample_sequence_batch(
            model,
            inference_strategy,
            context_tokens_tensor,
            context_length_tensor,
            audio_signal,
            audio_signal_length,
            tokens_to_generate,
            all_probs,
            compute_attention_mask=compute_attention_mask,
            compute_logprob=compute_logprob,
            temperature=temperature,
            end_strings=end_strings,
            extra={
                "top_p": top_p,
                "top_k": top_k,
                "greedy": greedy,
                "repetition_penalty": repetition_penalty,
                "min_tokens_to_generate": min_tokens_to_generate,
            },
            num_audios=num_audios,
            context_start_idx=context_start_idx,
        )

    for tokens, lengths, output_logits, full_logits, audio_feat_lens in batch_token_iterator:
        context_length += 1
    context_length += audio_feat_lens.min().item()
    if parallel_state.is_pipeline_last_stage():
        src = parallel_state.get_pipeline_model_parallel_last_rank()
        group = parallel_state.get_embedding_group()
        if compute_logprob:
            torch.distributed.broadcast(output_logits, src, group)
        if all_probs:
            src = parallel_state.get_pipeline_model_parallel_last_rank()
            group = parallel_state.get_embedding_group()
            torch.distributed.broadcast(full_logits, src, group)

    else:
        if parallel_state.is_pipeline_first_stage():
            src = parallel_state.get_pipeline_model_parallel_last_rank()
            group = parallel_state.get_embedding_group()

            if compute_logprob:
                precision = model._trainer.precision
                if precision in [16, "16"]:
                    dtype = torch.float16
                elif precision == "bf16":
                    dtype = torch.bfloat16
                else:
                    dtype = torch.float32
                output_logits = torch.empty(
                    tokens.size(0), context_length - 1, dtype=dtype, device=torch.device("cuda")
                )
                torch.distributed.broadcast(output_logits, src, group)

            if all_probs:
                src = parallel_state.get_pipeline_model_parallel_last_rank()
                group = parallel_state.get_embedding_group()
                full_logits = torch.empty(
                    tokens.size(0),
                    context_length - 1,
                    model.padded_vocab_size,
                    dtype=dtype,
                    device=torch.device("cuda"),
                )
                torch.distributed.broadcast(full_logits, src, group)
    if tokens is not None:
        return tokens[:, :context_length], output_logits, full_logits, audio_feat_lens
    return None


def generate(
    model,
    inputs: Union[Tuple, List[str]],
    tokens_to_generate=0,
    all_probs=False,
    temperature=1.0,
    add_BOS=False,
    top_k=0,
    top_p=0.0,
    greedy=False,
    compute_attention_mask=True,
    compute_logprob=False,
    repetition_penalty=1.0,
    end_strings=['<|endoftext|>'],
    min_tokens_to_generate=0,
    **strategy_args,
) -> OutputType:
    """
    Args:
        model (NLPModel): text generative model
        inputs (Union[tuple, List[str]]): if it is a tuple, it is assumed to be (context_tokens_tensor, context_length_tensor). Otherwise it it a list of prompt text strings
        tokens_to_generate (int): The maximum length of the tokens to be generated.
        all_probs (bool): Return the log prob for all the tokens
        temperature (float): sampling temperature
        add_BOS (bool): add the bos token at the begining of the prompt
        top_k (int): The number of highest probability vocabulary tokens to keep for top-k-filtering.
        top_p (float): If set to float < 1, only the most probable tokens with probabilities that add up to top_p or higher are kept for generation.
        greedy (bool):  Whether or not to use sampling ; use greedy decoding otherwise
        repetition_penalty (float): The parameter for repetition penalty. 1.0 means no penalty
        min_tokens_to_generate (int): The minimum length of the tokens to be generated
        strategy_args, the extra arguments are treated as inference strategy arguments
        end_strings, a list of strings to stop generation when they are encountered in the output.
    Returns:
        OutputType: It generates the output in a dictionary type. It has the following keys:
            sentences: List[str], output sentences
            tokens: List[List[str]], output sentences borken into tokens
            logprob: List[Tensor], log prob of generated tokens
            full_logprob: List[Tensor], log prob of all the tokens in the vocab
            token_ids: List[Tensor], output sentence token ids
            offsets: List[List[int]]  # list of tokens start positions in text
    """
    if 'strategy' in strategy_args:
        inference_strategy = strategy_args['strategy']
    else:
        inference_strategy = model_inference_strategy_dispatcher(model)
    tokenizer = model.tokenizer
    has_multi_audios = False
    num_audios = None
    context_start_idx = None
    audio_signal, audio_signal_length = None, None
    if isinstance(inputs, tuple) and len(inputs) == 2:
        context_tokens_tensor, context_length_tensor = inputs
    elif isinstance(inputs, tuple) and len(inputs) == 4:
        context_tokens_tensor, context_length_tensor, audio_signal, audio_signal_length = inputs
    elif isinstance(inputs, tuple) and len(inputs) == 6:  # multi-audio
        has_multi_audios = True
        (
            context_tokens_tensor,
            context_length_tensor,
            audio_signal,
            audio_signal_length,
            num_audios,
            context_start_idx,
        ) = inputs
    else:
        context_tokens_tensor, context_length_tensor = inference_strategy.tokenize_batch(
            inputs, tokens_to_generate, add_BOS
        )
    """  to unblock TP inference
    if torch.distributed.get_rank() == text_generation_utils.get_model_parallel_src_rank():
        if isinstance(inputs, tuple) and len(inputs) == 2:
            context_tokens_tensor, context_length_tensor = inputs
        elif isinstance(inputs, tuple) and len(inputs) == 4:
            context_tokens_tensor, context_length_tensor, audio_signal, audio_signal_length = inputs
        elif isinstance(inputs, tuple) and len(inputs) == 6:  # multi-audio
            has_multi_audios = True
            (
                context_tokens_tensor,
                context_length_tensor,
                audio_signal,
                audio_signal_length,
                num_audios,
                context_start_idx,
            ) = inputs
        else:
            context_tokens_tensor, context_length_tensor = inference_strategy.tokenize_batch(
                inputs, tokens_to_generate, add_BOS
            )

        send_generate_info(
            context_tokens_tensor,
            context_length_tensor,
            audio_signal,
            audio_signal_length,
            tokens_to_generate,
            all_probs,
            compute_logprob,
            temperature,
            top_k,
            top_p,
            greedy,
            repetition_penalty,
            min_tokens_to_generate,
            end_strings,
            num_audios,
            context_start_idx,
        )
    else:
        (
            context_length_tensor,
            context_tokens_tensor,
            audio_signal,
            audio_signal_length,
            tokens_to_generate,
            all_probs,
            compute_logprob,
            temperature,
            top_k,
            top_p,
            greedy,
            repetition_penalty,
            min_tokens_to_generate,
            end_strings,
            num_audios,
            context_start_idx,
        ) = receive_generate_info(has_multi_audios)
    """

    output = synced_generate(
        model,
        inference_strategy,
        context_tokens_tensor,
        context_length_tensor,
        audio_signal,
        audio_signal_length,
        tokens_to_generate,
        all_probs,
        temperature,
        compute_attention_mask=compute_attention_mask,
        compute_logprob=compute_logprob,
        top_k=top_k,
        top_p=top_p,
        greedy=greedy,
        repetition_penalty=repetition_penalty,
        end_strings=end_strings,
        min_tokens_to_generate=min_tokens_to_generate,
        num_audios=num_audios,
        context_start_idx=context_start_idx,
    )
    special_tokens = set()
    if hasattr(tokenizer, 'pad_token') and tokenizer.pad_token is not None:
        special_tokens.add(tokenizer.pad_token)
    if hasattr(tokenizer, 'eos_token') and tokenizer.eos_token is not None:
        special_tokens.add(tokenizer.eos_token)
    if hasattr(tokenizer, 'bos_token') and tokenizer.bos_token is not None:
        special_tokens.add(tokenizer.bos_token)
    if hasattr(tokenizer, 'cls_token') and tokenizer.cls_token is not None:
        special_tokens.add(tokenizer.cls_token)
    if hasattr(tokenizer, 'unk_token') and tokenizer.unk_token is not None:
        special_tokens.add(tokenizer.unk_token)
    if hasattr(tokenizer, 'sep_token') and tokenizer.sep_token is not None:
        special_tokens.add(tokenizer.sep_token)
    if hasattr(tokenizer, 'mask_token') and tokenizer.mask_token is not None:
        special_tokens.add(tokenizer.mask_token)
    if output is not None:
        decode_tokens, output_logits, full_logits, audio_feat_lens = output
        resp_sentences = []
        resp_sentences_seg = []

        decode_tokens = decode_tokens.cpu().numpy().tolist()
        for decode_token in decode_tokens:
            sentence = tokenizer.ids_to_text(decode_token)
            resp_sentences.append(sentence)
            if not isinstance(tokenizer, TabularTokenizer):
                words = []
                for token in decode_token:
                    if not isinstance(token, Iterable):
                        token = [token]
                    word = tokenizer.ids_to_tokens(token)
                    if isinstance(word, Iterable):
                        word = word[0]
                    if hasattr(tokenizer.tokenizer, 'byte_decoder'):
                        word = bytearray([tokenizer.tokenizer.byte_decoder[c] for c in word]).decode(
                            'utf-8', errors='replace'
                        )
                    words.append(word)
                resp_sentences_seg.append(words)
            else:
                words = tokenizer.text_to_tokens(sentence)
                resp_sentences_seg.append(words)

        # offsets calculation
        all_offsets = []
        for item in resp_sentences_seg:
            offsets = [0]
            for index, token in enumerate(item):
                if index != len(item) - 1:
                    if token in special_tokens:
                        offsets.append(offsets[-1])
                    else:
                        offsets.append(len(token) + offsets[-1])
            all_offsets.append(offsets)

        output = {}
        output['sentences'] = resp_sentences
        output['tokens'] = resp_sentences_seg
        output['logprob'] = output_logits
        output['full_logprob'] = full_logits
        output['token_ids'] = decode_tokens
        output['offsets'] = all_offsets
        output['audio_feat_lens'] = audio_feat_lens
        output = inference_strategy.post_generation_process(output)
        return output
    return None


def switch(val1, val2, boolean):
    boolean = boolean.type_as(val1)
    return (1 - boolean) * val1 + boolean * val2


def sample_sequence_batch(
    model,
    inference_strategy,
    context_tokens,
    context_lengths,
    audio_signal,
    audio_signal_length,
    tokens_to_generate,
    all_probs=False,
    compute_attention_mask=True,
    compute_logprob=False,
    type_ids=None,
    temperature=None,
    end_strings=['<|endoftext|>'],
    extra={},
    num_audios: Optional[torch.Tensor] = None,
    context_start_idx: Optional[List[List[int]]] = None,
):
    app_state = AppState()
    micro_batch_size = context_tokens.shape[0]
    reconfigure_num_microbatches_calculator(
        rank=app_state.global_rank,
        rampup_batch_size=None,
        global_batch_size=micro_batch_size,
        micro_batch_size=micro_batch_size,
        data_parallel_size=1,
    )
    assert tokens_to_generate > 0, "tokens_to_generate should be > 0"
    assert (
        model.cfg.get('sequence_parallel', False) == False
    ), 'sequence_parallel should be False during inference. Disable it in the model config if restoring from nemo or in hparams.yaml if restoring from PTL checkpoint'
    assert (
        model.cfg.get('activations_checkpoint_granularity', None) is None
    ), 'activations_checkpoint_granularity should be None during inference. Disable it in the model config if restoring from nemo or in hparams.yaml if restoring from PTL checkpoint'
    assert (
        model.cfg.get('activations_checkpoint_method', None) is None
    ), 'activations_checkpoint_method should be None during inference. Disable it in the model config if restoring from nemo or in hparams.yaml if restoring from PTL checkpoint'

    tokenizer = model.tokenizer
    # initialize the batch
    with torch.no_grad():
        context_tokens, input_embeddings, audio_feat_lens = inference_strategy.init_batch(
            context_tokens,
            context_lengths,
            audio_signal,
            audio_signal_length,
            compute_attention_mask,
            num_audios,
            context_start_idx,
        )
        audio_text_context_lengths = context_lengths + audio_feat_lens
        context_length = audio_text_context_lengths.min().item()
        # added eos_id to support the function generate_samples_eval that passes
        # eos_id as an argument and needs termination when that id id found.
        eod_id = tokenizer.eos_id
        counter = 0
        batch_size = context_tokens.size(0)
        is_done = torch.zeros([batch_size]).byte().cuda()
        tokens = context_tokens
        output_logits = None
        all_generated_indices = None  # used to track all generated indices
        # Generate enough tokens for the longest sequence
        maxlen = tokens_to_generate + audio_text_context_lengths.max().item()
        maxlen = inference_strategy.clip_max_len(maxlen)
        lengths = torch.ones([batch_size]).long().cuda() * maxlen
        while context_length < maxlen:
            batch, tensor_shape = inference_strategy.prepare_batch_at_step(
                tokens,
                input_embeddings,
                maxlen,
                micro_batch_size,
                counter,
                audio_text_context_lengths,
                context_length,
                compute_attention_mask,
            )
            output = inference_strategy.forward_step(batch, tensor_shape)
            if parallel_state.is_pipeline_last_stage():
                if compute_logprob:
                    output = output[0]['logits']
                    output = tensor_parallel.gather_from_tensor_model_parallel_region(output)
                    assert output is not None
                    logits = output[:, -1].view(batch_size, -1).contiguous()

                else:
                    logits = output[0]['logits'][:, -1].contiguous()
                    logits = tensor_parallel.gather_from_tensor_model_parallel_region(logits)
                    assert logits is not None
                    logits = logits.view(batch_size, -1)

                # make sure it will generate at least min_length
                min_length = extra.get('min_tokens_to_generate', 0)
                if min_length > 0:
                    within_min_length = (context_length - audio_text_context_lengths) < min_length
                    logits[within_min_length, eod_id] = -float('Inf')
                # make sure it won't sample outside the vocab_size range
                logits[:, tokenizer.vocab_size :] = -float('Inf')

                # started indicates whether the current token step passes the context_length, so we make sure not to overwrite the context tokens
                started = audio_text_context_lengths <= context_length
                if extra.get('greedy', False):
                    prev = torch.argmax(logits, dim=-1).view(-1)
                else:
                    logits = logits.float()
                    logits /= temperature
                    # handle repetition penality
                    logits = text_generation_utils.repetition_penalty(
                        logits, extra.get('repetition_penalty', 1.2), all_generated_indices
                    )
                    logits = text_generation_utils.top_k_logits(
                        logits, top_k=extra.get('top_k', 0), top_p=extra.get('top_p', 0.9), started=started
                    )
                    probs = F.softmax(logits, dim=-1)
                    # TODO(zhehuai)
                    probs = probs.nan_to_num(1.0)
                    prev = torch.multinomial(probs, num_samples=1).view(-1)

                # Clamp the predicted out of vocabulary tokens
                prev = torch.clamp(prev, max=tokenizer.vocab_size - 1)
                new_tokens = switch(tokens[:, context_length].view(-1), prev, started)

                # Replace sampled tokens w/ done token if EOD has already been sampled
                new_tokens = switch(new_tokens, eod_id, is_done)

                # post process the inference tokens based on the strategy
                inference_strategy.post_process(tokens, new_tokens, context_length)

                # Insert either new predicted or next prompt token
                tokens[:, context_length] = new_tokens

                if compute_logprob:
                    if output_logits is None:
                        output = F.log_softmax(output[:, :context_length, :], 2)

                        indices = torch.unsqueeze(tokens[:, 1 : context_length + 1], 2)
                        output_logits = torch.gather(output, 2, indices).squeeze(2)
                        all_generated_indices = indices[:, :, 0]
                        if all_probs:
                            full_logits = output
                    else:
                        output = F.log_softmax(output, 2)
                        indices = torch.unsqueeze(new_tokens, 1).unsqueeze(2)
                        new_output_logits = torch.gather(output, 2, indices).squeeze(2)

                        # TODO(rprenger) we're copying output_logits every time.  Should pre-allocate
                        output_logits = torch.cat([output_logits, new_output_logits], 1)
                        all_generated_indices = torch.cat([all_generated_indices, indices[:, :, 0]], 1)
                        if all_probs:
                            full_logits = torch.cat([full_logits, output], 1)

                src = parallel_state.get_pipeline_model_parallel_last_rank()
                group = parallel_state.get_embedding_group()
                torch.distributed.broadcast(new_tokens, src, group)

                #                done_token = (prev == eod_id).byte() & started.byte()
                done_token = inference_strategy.end_of_generation_condition(
                    tokens[:, : context_length + 1], prev, eod_id, end_strings
                )
                done_token = done_token.byte() & started.byte()

                just_finished = (done_token & ~is_done).bool()
                lengths[just_finished.view(-1)] = context_length
                is_done = is_done | done_token

                done = torch.all(is_done)
                src = parallel_state.get_pipeline_model_parallel_last_rank()
                group = parallel_state.get_pipeline_model_parallel_group()
                torch.distributed.broadcast(done, src, group)
                if compute_logprob:
                    if all_probs:
                        yield tokens, lengths, output_logits, full_logits, audio_feat_lens
                    else:
                        yield tokens, lengths, output_logits, None, audio_feat_lens
                else:
                    yield tokens, lengths, None, None, audio_feat_lens

            else:
                if parallel_state.is_pipeline_first_stage():
                    src = parallel_state.get_pipeline_model_parallel_last_rank()
                    group = parallel_state.get_embedding_group()
                    new_tokens = torch.empty_like(tokens[:, context_length])
                    torch.distributed.broadcast(new_tokens, src, group)
                    tokens[:, context_length] = new_tokens
                    yield tokens, None, None, None, audio_feat_lens
                else:
                    yield None, None, None, None, audio_feat_lens

                done = torch.cuda.ByteTensor([0])
                src = parallel_state.get_pipeline_model_parallel_last_rank()
                group = parallel_state.get_pipeline_model_parallel_group()
                torch.distributed.broadcast(done, src, group)

            context_length += 1
            counter += 1
            if done:
                break



# Copyright (c) 2021, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
import random
from typing import Dict, List

import torch
from omegaconf import DictConfig
from omegaconf.dictconfig import DictConfig
from torch import nn
from torch.nn import functional as F

from nemo.collections.asr.modules.common.transformer.transformer_encoders_nlp import TransformerEncoder
from nemo.collections.common.parts import form_attention_mask, transformer_weights_init
from nemo.core.classes.module import NeuralModule
from nemo.core.neural_types import AcousticEncodedRepresentation, AudioSignal, LengthsType, NeuralType, SpectrogramType


class TransposeLast(torch.nn.Module):
    """
    Transposes last dimension. Useful for adding to a sequential block.
    """

    def forward(self, x):
        return x.transpose(-2, -1)


class SamePad(torch.nn.Module):
    def __init__(self, kernel_size):
        super().__init__()
        self.remove = kernel_size % 2 == 0

    def forward(self, x):
        if self.remove:
            x = x[:, :, :-1]
        return x


class ConvFeatureEncoder(NeuralModule):
    """
    Encoder used to isolate features in raw audio for Wav2Vec style training.
    Treated as preprocessor module in NeMo ASR training. Defaults values are
    for base model found in Baeski et al (https://arxiv.org/abs/2006.11477),
    save for use of layer normalization as default schema. (Chosen for stability.)
    """

    @property
    def input_types(self):
        """Returns definitions of module input ports.
        input_signal:
            0: AxisType(BatchTag)
            1: AxisType(TimeTag)
        input_signal_length:
            0: AxisType(BatchTag)
        Note: length is in number of samples, not seconds
        """
        return {
            "input_signal": NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        """Returns definitions of module output ports.
        For compatibility, processed features are treated as Spectrogram types
        processed_signal:
            0: AxisType(BatchTag)
            1: AxisType(ChannelTag)
            2: AxisType(ProcessedTimeTag)
        processed_signal_length:
            0: AxisType(BatchTag)
        """
        return {
            "processed_signal": NeuralType(('B', 'C', 'T'), SpectrogramType()),
            "processed_signal_length": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(
        self,
        conv_layers: List[Dict[str, int]],
        extractor_mode: str = "layer_norm",
        conv_bias: bool = False,
        feature_grad_mult=1.0,
        normalize_audio=True,
        embedding_dim=768,
    ):
        super().__init__()

        self.grad_mult = feature_grad_mult
        self.normalize_input = normalize_audio

        def block(
            n_in,
            n_out,
            k,
            stride,
            is_layer_norm=False,
            is_group_norm=False,
            conv_bias=False,
        ):
            def make_conv():
                conv = nn.Conv1d(n_in, n_out, k, stride=stride, bias=conv_bias)
                nn.init.kaiming_normal_(conv.weight)
                return conv

            assert (is_layer_norm and is_group_norm) is False, "layer norm and group norm are exclusive"

            if is_layer_norm:
                return nn.Sequential(
                    make_conv(),
                    nn.Sequential(TransposeLast(), nn.LayerNorm(dim, elementwise_affine=True), TransposeLast()),
                    nn.GELU(),
                )
            elif is_group_norm:
                return nn.Sequential(
                    make_conv(),
                    nn.GroupNorm(dim, dim, affine=True),
                    nn.GELU(),
                )
            else:
                return nn.Sequential(make_conv(), nn.GELU())

        in_d = 1
        self.layer_cfg = conv_layers
        self.conv_layers = nn.ModuleList()
        self.mode = extractor_mode
        for i, cl in enumerate(conv_layers):
            assert len(cl) == 3, "invalid conv definition: " + str(cl)
            dim, k, stride = cl["emb_dim"], cl["kernel_size"], cl["stride"]

            self.conv_layers.append(
                block(
                    in_d,
                    dim,
                    k,
                    stride,
                    is_layer_norm=self.mode == "layer_norm",
                    is_group_norm=self.mode == "group_norm" and i == 0,  # applied to first layer only
                    conv_bias=conv_bias,
                )
            )
            in_d = dim

        # Model Layers
        final_conv_dim = self.layer_cfg[-1]["emb_dim"]  # Select last conv output layer dimension
        self.post_extract_proj = (  # To project feature encodings to transformer
            nn.Linear(final_conv_dim, embedding_dim) if final_conv_dim != embedding_dim else None
        )
        self.layer_norm = nn.LayerNorm(embedding_dim)

    def apply_layers(self, x):
        for conv in self.conv_layers:
            x = conv(x)
        return x

    def normalize(self, source, lengths):
        with torch.no_grad():  # Normalizes audio source
            for i in range(lengths.size(0)):
                orig = source[i, : lengths[i]]
                norm = F.layer_norm(orig, orig.shape)
                source[i, : lengths[i]] = norm
        return source

    def forward(self, input_signal, length):
        if self.normalize_input:
            input_signal = self.normalize(input_signal, length)

        # BxT -> BxCxT
        processed_signal = input_signal.unsqueeze(1)

        # Applies grad mult scaling
        if self.grad_mult > 0:
            processed_signal = self.apply_layers(processed_signal)
            if self.grad_mult != 1.0:
                processed_signal = GradMultiply.apply(processed_signal, self.grad_mult)
        else:
            with torch.no_grad():  # 0 indicates frozen feature encoder
                processed_signal = self.apply_layers(processed_signal)

        processed_signal = processed_signal.transpose(1, 2)  # B,T,C
        # Project to embedding
        if self.post_extract_proj is not None:
            processed_signal = self.post_extract_proj(processed_signal)

        # Adding normalization for output
        if self.mode == "layer_norm":
            processed_signal = self.layer_norm(processed_signal)

        processed_signal = processed_signal.transpose(1, 2)  # B,C,T

        # Feature lengths will have been changed through convolutions
        processed_signal_length = self.get_lengths(audio_lengths=length)

        return processed_signal, processed_signal_length

    def get_lengths(self, audio_lengths):
        # converts audio lengths to timestep lengths
        for conv in self.layer_cfg:
            kernel = conv["kernel_size"]
            stride = conv["stride"]
            audio_lengths = (
                torch.div(audio_lengths - kernel, stride, rounding_mode='floor') + 1
            )  # from pytorch documentation
        return audio_lengths


class Wav2VecTransformerEncoder(TransformerEncoder):
    """
    Encoder module following Transformer encoder paradigm
    as described in Vaswani et al. (https://arxiv.org/abs/1706.03762). Used for Wav2Vec
    style encoding of context vectors as described by in Baeski et al (https://arxiv.org/abs/2006.11477).
    Takes convolutional encodings of all time steps and adds to features before applying series
    of self-attention layers.

    Example configs may be found at: https://github.com/NVIDIA/NeMo/tree/main/examples/asr/conf/wav2vec

    Args:
            layer_drop: Floating point value specifying proportion of module for layer dropout (See Fan et al. https://arxiv.org/pdf/1909.11556.pdf).
                    If non-zero, each layer will draw from uniform probability to determine if applied in current forward call.
                    Occurs only during training step
            pos_embed: Config specifying parameters for contextual embedding convolutions. Module configures convolutional padding
                    to maintain number of time steps
                    Must contain following:
                            embedding_dim: Depth/number of channels of each time step from feature encoding
                            conv_pos: Kernel size for convolution
                            conv_pos_groups: Number of groups for convolution
            transformer: Config for transformer encoder. Uses self-attention layers found in: nemo.collections.nlp.modules.common.transformer
                    Must contain followign:
                            num_layers: Number of attention layers
                            hidden_size: Expected input depth (embedding size between model layers)
                            inner_size: Depth of embeddings within feed-forward sections of encoder layers
                            num_attention_heads: Number of attention heads
                            attn_score_dropout: Probability of dropout applied to attention scores
                            attn_layer_dropout: Probability of dropout applied to the output of the attention layers (prior to normalization)
                            ffn_dropout: Probability of dropout applied to feed-forward modules
                            hidden_act: Activation function for hidden layers
    """

    def __init__(self, pos_embed: DictConfig, transformer: DictConfig, layer_drop: float = 0.0):
        super().__init__(**transformer)  # see nlp.collections

        # positional convolutional embeddings
        emb_dim = pos_embed.embedding_dim
        self.pos_conv = nn.Conv1d(
            emb_dim,
            emb_dim,
            kernel_size=pos_embed.conv_pos,
            padding=pos_embed.conv_pos // 2,  # Padding size preserves time step length
            groups=pos_embed.conv_pos_groups,
        )

        self.layer_drop = layer_drop

        self.dropout = transformer.attn_layer_dropout  # He initialization
        std = math.sqrt((4 * (1.0 - self.dropout)) / (pos_embed.conv_pos * pos_embed.embedding_dim))
        nn.init.normal_(self.pos_conv.weight, mean=0, std=std)
        nn.init.constant_(self.pos_conv.bias, 0)

        self.pos_conv = nn.utils.weight_norm(self.pos_conv, name="weight", dim=2)
        self.pos_conv = nn.Sequential(self.pos_conv, SamePad(pos_embed.conv_pos), nn.GELU())

        self.layer_norm = nn.LayerNorm(emb_dim)
        self.apply(lambda x: transformer_weights_init(x, xavier=False))

    @property
    def input_types(self):
        """Returns definitions of module output ports.
        We treat features as SpectrogramType for Nemo compatibility
        audio_signal:
            0: AxisType(BatchTag)
            1: AxisType(ChannelTag)
            2: AxisType(ProcessedTimeTag)
        length:
            0: AxisType(BatchTag)
        """
        return {
            "audio_signal": NeuralType(('B', 'C', 'T'), SpectrogramType()),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    def output_types(self):
        """Returns definitions of module output ports.
        We're using SpectrogramType for now to keep things Nemo safe
        processed_signal:
            0: AxisType(BatchTag)
            1: AxisType(ChannelTag)
            2: AxisType(ProcessedTimeTag)
        processed_length:
            0: AxisType(BatchTag)
        """
        return {
            "processed_signal": NeuralType(('B', 'C', 'T'), AcousticEncodedRepresentation()),
            "processed_length": NeuralType(tuple('B'), LengthsType()),
        }

    def forward(self, audio_signal, length):

        # Padding mask needed for transformer
        padding_mask = self.create_padding_mask(length)

        # Applying padding before convolution
        for idx, len in enumerate(length):
            audio_signal[idx, :, len:] = 0.0

        signal_conv = self.pos_conv(audio_signal)  # B, C, T
        audio_signal = audio_signal + signal_conv

        audio_signal = audio_signal.transpose(1, 2)  # B, C, T -> B, T, C
        audio_signal = self.layer_norm(audio_signal)

        context_emb = self.apply_transformer(audio_signal, padding_mask=padding_mask)

        context_emb = context_emb.transpose(1, 2)  # B, T, C -> B, C, T

        return context_emb, length  # Returning length for NeMo compatibility

    def apply_transformer(self, x, padding_mask=None):
        encoder_attn_mask = form_attention_mask(padding_mask)
        if (
            self.layer_drop and self.training
        ):  # Stochastic layer drop as in: Huang et al. https://arxiv.org/pdf/1603.09382.pdf
            for _, layer in enumerate(self.layers):
                p = random.random()
                if p > self.layer_drop:
                    x = layer(x, encoder_attn_mask, x)
        else:
            for _, layer in enumerate(self.layers):
                x = layer(x, encoder_attn_mask, x)
        return x

    def create_padding_mask(self, length):
        # Broadcast to vectorize creating the padding mask
        max_len = max(length)
        padding_mask = torch.arange(max_len, device=length.device)

        # Switch to binary for transformer, 1 for valid tokens, 0 for padding
        padding_mask = (padding_mask.expand(len(length), max_len) < length.unsqueeze(1)).type(torch.uint8)

        return padding_mask


class GradMultiply(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, scale):
        ctx.scale = scale
        res = x.new(x)
        return res

    @staticmethod
    def backward(ctx, grad):
        return grad * ctx.scale, None



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nemo.collections.avlm.data.energon.avlm_sample_config import AVLMSampleConfig
from nemo.collections.avlm.data.energon.avlm_task_encoder import AVLMTaskEncoder
from nemo.collections.multimodal.data.energon import EnergonMultiModalDataModule


class AVLMDataModule(EnergonMultiModalDataModule):
    """
    Energon Data module for media(audio and/or image and/or video and text)-to-text LLM.

    It looks for a path leads to path/.nv-meta/dataset.yaml
    which looks like the follows:
    sample_type:
      __module__: nemo.collections.avlm.data.energon.media_to_text_config
      __class__: MediaToMediaSample
    field_map:
      image: image_file_extension
      video: video_file_extension
      audio: audio_file_extension
      offset: ...

    More data preparation details can be found at:
    https://github.com/NVIDIA/Megatron-Energon/blob/24123d4b63a451980eec6ee0ddaa4f007ef562e7/docs/source/basic/data_prep.md?plain=1#L85
    """

    def __init__(
        self,
        path: str,
        tokenizer=None,
        audio_processor=None,
        image_processor=None,
        seq_length: int = 2048,
        micro_batch_size: int = 1,
        global_batch_size: int = 1,
        num_workers: int = 1,
        num_val_workers: int | None = None,
        pin_memory: bool = True,
        shuffle_buffer_size: int = 100,
        max_samples_per_sequence: int | None = None,
        multimodal_sample_config: Optional[AVLMSampleConfig] = AVLMSampleConfig(),
        task_encoder: Optional[AVLMTaskEncoder] = None,
        decoder_seq_length: Optional[int] = None,
        packing_buffer_size: Optional[int] = None,
        validation_task_encoder: Optional[AVLMTaskEncoder] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            path,
            tokenizer,
            image_processor,
            seq_length,
            micro_batch_size,
            global_batch_size,
            num_workers,
            num_val_workers,
            pin_memory,
            shuffle_buffer_size,
            max_samples_per_sequence,
            multimodal_sample_config,
            task_encoder,
            decoder_seq_length,
            packing_buffer_size,
            validation_task_encoder,
            **kwargs,
        )
        self.audio_processor = audio_processor

        # audio, video and image files wil be passed to the task encoder as raw
        # bytes which will be processed in a more fine-grained way.
        self.kwargs["auto_decode"] = False



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import yaml

from nemo.collections.common.metrics.perf_metrics import FLOPsMeasurementCallback

LLAMA2_CFG_STR = """
    run:
        name: train_llama2_7b_tp1_pp1_FP8_1node_15steps
    trainer:
        num_nodes: 1
        devices: 8
        precision: bf16
    exp_manager:
        explicit_log_dir: "results/logs"
    model:
        micro_batch_size: 1
        global_batch_size: 128
        encoder_seq_length: 4096
        max_position_embeddings: 4096
        num_layers: 32
        hidden_size: 4096
        ffn_hidden_size: 11008
        num_attention_heads: 32
"""

NEMOTRON_CFG_STR = """
    run:
      name: train_nemotron_8b_tp2_pp1_FP8_8node_20steps
    trainer:
      num_nodes: 8
      devices: 8
      precision: bf16
    exp_manager:
      explicit_log_dir: null
    model:
      micro_batch_size: 4
      global_batch_size: 256
      encoder_seq_length: 4096
      max_position_embeddings: 4096
      num_layers: 32
      hidden_size: 4096
      ffn_hidden_size: 16384
      num_attention_heads: 32
      fp8: true
"""

UNSUPPORTED_MODEL_CFG_STR = """
    run:
        name: unsupported_model
    trainer:
        num_nodes: 1
        devices: 8
        precision: bf64
    exp_manager:
        explicit_log_dir: null
    model:
        micro_batch_size: 1
        global_batch_size: 1
        encoder_seq_length: 1
        max_position_embeddings: 1
        num_layers: 1
        hidden_size: 1
        ffn_hidden_size: 1
        num_attention_heads: 1
"""

NULL_MODEL_CFG_STR = """
    run:
        name: null
"""


@pytest.fixture
def model_config(cfg):
    return yaml.safe_load(cfg)


@pytest.mark.unit
@pytest.mark.parametrize(
    "cfg, model_name, train_step_time, expected_value",
    [
        (LLAMA2_CFG_STR, None, 8, 351.14),
        (LLAMA2_CFG_STR, "llama2", 8, 351.14),
        (LLAMA2_CFG_STR, None, [8, 8, 8, 8], 351.14),
        (NEMOTRON_CFG_STR, None, 1.31, 602.42),
        (
            UNSUPPORTED_MODEL_CFG_STR,
            None,
            1,  # model_name in config is unsupported
            "Failed to extract valid model name from or missing FLOPs calculations for unsupported_model",
        ),
        (
            UNSUPPORTED_MODEL_CFG_STR,
            "unknown_model",
            1,  # overrided model name is unsupported
            "Failed to extract valid model name from or missing FLOPs calculations for unknown_model",
        ),
        (
            NULL_MODEL_CFG_STR,
            None,
            1,  # both- config and overrided model name are None
            "Failed to extract valid model name from or missing FLOPs calculations for None",
        ),
    ],
)
def test_eval_tflops_per_sec_per_gpu(model_config, model_name, train_step_time, expected_value):
    if isinstance(expected_value, (int, float)):
        flops_callback = FLOPsMeasurementCallback(model_config, model_name=model_name)
        tflops_per_sec_per_gpu = flops_callback.eval_tflops_per_sec_per_gpu(train_step_time)
        assert tflops_per_sec_per_gpu == pytest.approx(expected_value, rel=1e-4)

        if model_name is None:
            # extract valid model name with delimiter='-'
            model_config["run"]["name"] = model_config["run"]["name"].replace("_", ".")
            flops_callback = FLOPsMeasurementCallback(model_config, model_name=model_name)
            tflops_per_sec_per_gpu = flops_callback.eval_tflops_per_sec_per_gpu(train_step_time)
            assert tflops_per_sec_per_gpu == pytest.approx(expected_value, rel=1e-4)

            # # extract valid model name from a string
            model_config["run"]["name"] = model_config["run"]["name"].replace("_", "")
            flops_callback = FLOPsMeasurementCallback(model_config, model_name=model_name)
            tflops_per_sec_per_gpu = flops_callback.eval_tflops_per_sec_per_gpu(train_step_time)
            assert tflops_per_sec_per_gpu == pytest.approx(expected_value, rel=1e-4)

    if isinstance(expected_value, str):
        flops_callback = FLOPsMeasurementCallback(model_config, model_name=model_name)
        with pytest.raises(KeyError, match=expected_value):
            _ = flops_callback.eval_tflops_per_sec_per_gpu(train_step_time)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Optional, Tuple

import torch
import torch.distributed
from megatron.core.transformer import TransformerConfig
from torch import Tensor
from transformers import MllamaConfig as HFMllamaConfig
from transformers import MllamaForConditionalGeneration
from transformers.models.mllama.configuration_mllama import MllamaTextConfig, MllamaVisionConfig

from nemo.collections.common.tokenizers.huggingface.auto_tokenizer import AutoTokenizer
from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.collections.vlm.mllama.model.base import (
    CrossAttentionTextConfig,
    CrossAttentionVisionConfig,
    MLlamaModel,
    MLlamaModelConfig,
)
from nemo.export.trt_llm.nemo_ckpt_loader.nemo_file import load_distributed_model_weights
from nemo.lightning import io, teardown
from nemo.lightning.io.state import _ModelState
from nemo.lightning.pytorch.utils import dtype_from_hf
from nemo.utils import logging

# pylint: disable=C0115,C0116,C0301


@dataclass
class MLlamaConfig11B(MLlamaModelConfig):
    language_model_config: Optional[TransformerConfig] = field(default_factory=lambda: CrossAttentionTextConfig())
    vision_model_config: Optional[TransformerConfig] = field(
        default_factory=lambda: CrossAttentionVisionConfig(vision_chunk_size=448)
    )


@dataclass
class MLlamaConfig11BInstruct(MLlamaModelConfig):
    language_model_config: Optional[TransformerConfig] = field(default_factory=lambda: CrossAttentionTextConfig())
    vision_model_config: Optional[TransformerConfig] = field(
        default_factory=lambda: CrossAttentionVisionConfig(vision_chunk_size=560)
    )


@dataclass
class MLlamaConfig90B(MLlamaModelConfig):
    language_model_config: Optional[TransformerConfig] = field(
        default_factory=lambda: CrossAttentionTextConfig(
            hidden_size=8192,
            ffn_hidden_size=28672,
            num_attention_heads=64,
            num_layers=80,
            num_cross_attention_layers=20,
        )
    )
    vision_model_config: Optional[TransformerConfig] = field(
        default_factory=lambda: CrossAttentionVisionConfig(vision_chunk_size=560, text_hidden_size=8192)
    )


@dataclass
class MLlamaConfig90BInstruct(MLlamaConfig90B):
    pass


@io.model_importer(MLlamaModel, "hf")
class HFMLlamaImporter(io.ModelConnector["MLlamaModel", MLlamaModel]):
    def init(self) -> MLlamaModel:
        return MLlamaModel(self.config, tokenizer=self.tokenizer)

    def local_path(self, base_path: Optional[Path] = None) -> Path:
        # note: this entire function is for debugging
        output_path = super().local_path(base_path)
        return output_path

    def apply(self, output_path: Path) -> Path:
        source = MllamaForConditionalGeneration.from_pretrained(str(self), torch_dtype="auto")

        state_dict = _rename_xattn_layer_nums_hf(source.state_dict())
        source = _ModelState(state_dict)
        target = self.init()
        trainer = self.nemo_setup(target)
        self.convert_state(source, target)
        self.nemo_save(output_path, trainer)

        print(f"Converted Mllama model to Nemo, model saved to {output_path}")

        teardown(trainer, target)
        del trainer, target

        return output_path

    def convert_state(self, source, target):
        mapping = {}
        transforms = []
        mapping.update(
            {
                "model.language_model.layers.*.self_attn.o_proj.weight": "language_model.decoder.layers.*.self_attention.linear_proj.weight",
                "model.language_model.xattn_layers.*.cross_attn.o_proj.weight": "language_model.decoder.xattn_layers.*.cross_attention.linear_proj.weight",
                "model.language_model.xattn_layers.*.cross_attn.q_proj.weight": "language_model.decoder.xattn_layers.*.cross_attention.linear_q.weight",
                "model.language_model.norm.weight": "language_model.decoder.final_layernorm.weight",
                "lm_head.weight": "language_model.output_layer.weight",
                "model.language_model.layers.*.post_attention_layernorm.weight": "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight",
                "model.language_model.layers.*.mlp.down_proj.weight": "language_model.decoder.layers.*.mlp.linear_fc2.weight",
                "model.language_model.layers.*.input_layernorm.weight": "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight",
                "model.language_model.xattn_layers.*.cross_attn.k_norm.weight": "language_model.decoder.xattn_layers.*.cross_attention.k_layernorm.weight",
                "model.language_model.xattn_layers.*.input_layernorm.weight": "language_model.decoder.xattn_layers.*.cross_attention.linear_q.layer_norm_weight",
                "model.language_model.xattn_layers.*.cross_attn.q_norm.weight": "language_model.decoder.xattn_layers.*.cross_attention.q_layernorm.weight",
                "model.language_model.xattn_layers.*.post_attention_layernorm.weight": "language_model.decoder.xattn_layers.*.mlp.linear_fc1.layer_norm_weight",
                "model.language_model.xattn_layers.*.mlp.down_proj.weight": "language_model.decoder.xattn_layers.*.mlp.linear_fc2.weight",
            }
        )

        transforms.extend(
            [
                io.state_transform(
                    source_key="model.language_model.xattn_layers.*.cross_attn_attn_gate",
                    target_key="language_model.decoder.xattn_layers.*.gate_attn",
                    fn=_import_gate,
                ),
                io.state_transform(
                    source_key="model.language_model.xattn_layers.*.cross_attn_mlp_gate",
                    target_key="language_model.decoder.xattn_layers.*.gate_ffn",
                    fn=_import_gate,
                ),
                io.state_transform(
                    source_key=(
                        "model.language_model.layers.*.self_attn.q_proj.weight",
                        "model.language_model.layers.*.self_attn.k_proj.weight",
                        "model.language_model.layers.*.self_attn.v_proj.weight",
                    ),
                    target_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
                    fn=_import_text_qkv,
                ),
                io.state_transform(
                    source_key=(
                        "model.language_model.layers.*.mlp.gate_proj.weight",
                        "model.language_model.layers.*.mlp.up_proj.weight",
                    ),
                    target_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
                    fn=_import_simple_concat,
                ),
                io.state_transform(
                    source_key=(
                        "model.language_model.xattn_layers.*.cross_attn.k_proj.weight",
                        "model.language_model.xattn_layers.*.cross_attn.v_proj.weight",
                    ),
                    target_key="language_model.decoder.xattn_layers.*.cross_attention.linear_kv.weight",
                    fn=_import_text_kv,
                ),
                io.state_transform(
                    source_key=(
                        "model.language_model.xattn_layers.*.mlp.gate_proj.weight",
                        "model.language_model.xattn_layers.*.mlp.up_proj.weight",
                    ),
                    target_key="language_model.decoder.xattn_layers.*.mlp.linear_fc1.weight",
                    fn=_import_simple_concat,
                ),
                io.state_transform(
                    source_key="model.language_model.embed_tokens.weight",
                    target_key=(
                        "language_model.embedding.word_embeddings.weight",
                        "language_model.learnable_embedding.weight",
                    ),
                    fn=_import_embedding_hf,
                ),
            ]
        )

        v = "vision_model.vision_encoder"
        mapping.update(
            {
                "model.vision_model.global_transformer.layers.*.self_attn.o_proj.weight": f"{v}.global_transformer.layers.*.self_attention.linear_proj.weight",
                "model.vision_model.global_transformer.layers.*.gate_attn": f"{v}.global_transformer.layers.*.gate_attn",
                "model.vision_model.global_transformer.layers.*.gate_ffn": f"{v}.global_transformer.layers.*.gate_ffn",
                "model.vision_model.global_transformer.layers.*.input_layernorm.bias": f"{v}.global_transformer.layers.*.input_layernorm.bias",
                "model.vision_model.global_transformer.layers.*.input_layernorm.weight": f"{v}.global_transformer.layers.*.input_layernorm.weight",
                "model.vision_model.global_transformer.layers.*.post_attention_layernorm.bias": f"{v}.global_transformer.layers.*.pre_mlp_layernorm.bias",
                "model.vision_model.global_transformer.layers.*.post_attention_layernorm.weight": f"{v}.global_transformer.layers.*.pre_mlp_layernorm.weight",
                "model.vision_model.global_transformer.layers.*.mlp.fc1.bias": f"{v}.global_transformer.layers.*.mlp.linear_fc1.bias",
                "model.vision_model.global_transformer.layers.*.mlp.fc1.weight": f"{v}.global_transformer.layers.*.mlp.linear_fc1.weight",
                "model.vision_model.global_transformer.layers.*.mlp.fc2.bias": f"{v}.global_transformer.layers.*.mlp.linear_fc2.bias",
                "model.vision_model.global_transformer.layers.*.mlp.fc2.weight": f"{v}.global_transformer.layers.*.mlp.linear_fc2.weight",
                "model.vision_model.transformer.layers.*.self_attn.o_proj.weight": f"{v}.transformer.layers.*.self_attention.linear_proj.weight",
                "model.vision_model.transformer.layers.*.input_layernorm.bias": f"{v}.transformer.layers.*.input_layernorm.bias",
                "model.vision_model.transformer.layers.*.input_layernorm.weight": f"{v}.transformer.layers.*.input_layernorm.weight",
                "model.vision_model.transformer.layers.*.post_attention_layernorm.bias": f"{v}.transformer.layers.*.pre_mlp_layernorm.bias",
                "model.vision_model.transformer.layers.*.post_attention_layernorm.weight": f"{v}.transformer.layers.*.pre_mlp_layernorm.weight",
                "model.vision_model.transformer.layers.*.mlp.fc1.bias": f"{v}.transformer.layers.*.mlp.linear_fc1.bias",
                "model.vision_model.transformer.layers.*.mlp.fc1.weight": f"{v}.transformer.layers.*.mlp.linear_fc1.weight",
                "model.vision_model.transformer.layers.*.mlp.fc2.bias": f"{v}.transformer.layers.*.mlp.linear_fc2.bias",
                "model.vision_model.transformer.layers.*.mlp.fc2.weight": f"{v}.transformer.layers.*.mlp.linear_fc2.weight",
                "model.vision_model.class_embedding": f"{v}.class_embedding",
                "model.vision_model.gated_positional_embedding.embedding": f"{v}.positional_embedding",
                "model.vision_model.gated_positional_embedding.tile_embedding.weight": f"{v}.gated_tile_positional_embedding.weight",
                "model.vision_model.gated_positional_embedding.gate": f"{v}.gated_positional_embedding_gate",
                "model.vision_model.layernorm_post.bias": f"{v}.ln_post.bias",
                "model.vision_model.layernorm_post.weight": f"{v}.ln_post.weight",
                "model.vision_model.layernorm_pre.bias": f"{v}.ln_pre.bias",
                "model.vision_model.layernorm_pre.weight": f"{v}.ln_pre.weight",
                "model.vision_model.post_tile_positional_embedding.embedding.weight": f"{v}.post_tile_pos_embed.embedding.weight",
                "model.vision_model.post_tile_positional_embedding.gate": f"{v}.post_tile_pos_embed.gate",
                "model.vision_model.pre_tile_positional_embedding.embedding.weight": f"{v}.pre_tile_pos_embed.embedding.weight",
                "model.vision_model.pre_tile_positional_embedding.gate": f"{v}.pre_tile_pos_embed.gate",
                "model.multi_modal_projector.bias": "vision_model.vision_projection.encoder.bias",
                "model.multi_modal_projector.weight": "vision_model.vision_projection.encoder.weight",
            }
        )
        transforms.extend(
            [
                io.state_transform(
                    source_key=(
                        "model.vision_model.global_transformer.layers.*.self_attn.q_proj.weight",
                        "model.vision_model.global_transformer.layers.*.self_attn.k_proj.weight",
                        "model.vision_model.global_transformer.layers.*.self_attn.v_proj.weight",
                    ),
                    target_key=(f"{v}.global_transformer.layers.*.self_attention.linear_qkv.weight"),
                    fn=_import_vision_qkv,
                ),
                io.state_transform(
                    source_key=(
                        "model.vision_model.transformer.layers.*.self_attn.q_proj.weight",
                        "model.vision_model.transformer.layers.*.self_attn.k_proj.weight",
                        "model.vision_model.transformer.layers.*.self_attn.v_proj.weight",
                    ),
                    target_key=(f"{v}.transformer.layers.*.self_attention.linear_qkv.weight"),
                    fn=_import_vision_qkv,
                ),
                io.state_transform(
                    source_key="model.vision_model.patch_embedding.weight",
                    target_key=f"{v}.conv1._linear.weight",
                    fn=_import_patch_embedding_hf,
                ),
            ]
        )

        return io.apply_transforms(source, target, mapping=mapping, transforms=transforms)

    @property
    def tokenizer(self) -> "AutoTokenizer":
        return AutoTokenizer(self.save_hf_tokenizer_assets(str(self)))

    @property
    def config(self) -> MLlamaModelConfig:
        from transformers import AutoConfig

        source = AutoConfig.from_pretrained(str(self))

        return MLlamaModelConfig(
            language_model_config=self._language_model_config(source),
            vision_model_config=self._vision_model_config(source),
        )

    def _language_model_config(self, source) -> Optional[CrossAttentionTextConfig]:
        def _calculate_num_layers(num_hidden_layers, cross_attention_layers):
            return num_hidden_layers - len(cross_attention_layers)

        return CrossAttentionTextConfig(
            rotary_base=source.text_config.rope_theta,
            seq_length=8192,
            num_layers=_calculate_num_layers(
                source.text_config.num_hidden_layers,
                source.text_config.cross_attention_layers,
            ),
            num_cross_attention_layers=len(source.text_config.cross_attention_layers),
            hidden_size=source.text_config.hidden_size,
            ffn_hidden_size=source.text_config.intermediate_size,
            num_attention_heads=source.text_config.num_attention_heads,
            num_query_groups=source.text_config.num_key_value_heads,
            vocab_size=source.text_config.vocab_size,
            fp16=(dtype_from_hf(source) == torch.float16),
            bf16=(dtype_from_hf(source) == torch.bfloat16),
            params_dtype=dtype_from_hf(source),
        )

    def _vision_model_config(self, source) -> Optional[CrossAttentionVisionConfig]:
        return CrossAttentionVisionConfig(
            num_layers=source.vision_config.num_hidden_layers,
            hidden_size=source.vision_config.hidden_size,
            num_attention_heads=source.vision_config.attention_heads,
            vision_chunk_size=source.vision_config.image_size,
            vision_max_num_chunks=source.vision_config.max_num_tiles,
            text_hidden_size=source.text_config.hidden_size,
            fp16=(dtype_from_hf(source) == torch.float16),
            bf16=(dtype_from_hf(source) == torch.bfloat16),
            params_dtype=dtype_from_hf(source),
        )


@io.model_exporter(MLlamaModel, "hf")
class HFMLlamaExporter(io.ModelConnector[MLlamaModel, "MllamaForConditionalGeneration"]):
    """
    Exporter class for converting NeMo MLlama model to HuggingFace format.

    Inherits:
        io.ModelConnector: Connector interface to handle setup, save, and load using the Lightning framework.

    Methods:
        init: Initializes a new HuggingFace MLlama model instance.
        apply: Converts the NeMo model to HuggingFace format and saves it.
        convert_state: Maps and transforms the state dictionary from NeMo to HuggingFace format.
        config: Generates and returns the HuggingFace MLlama config for the model.
    """

    def init(self, dtype=torch.bfloat16) -> "MllamaForConditionalGeneration":
        """
        Initializes a HuggingFace MllamaForConditionalGeneration model.

        Args:
            dtype: The data type to use for the model (default: torch.bfloat16)

        Returns:
            MllamaForConditionalGeneration: A HuggingFace MLlama model initialized with the configuration.
        """
        from transformers.modeling_utils import no_init_weights

        with no_init_weights():
            return MllamaForConditionalGeneration._from_config(self.config, torch_dtype=dtype)

    def apply(self, output_path: Path) -> Path:
        """
        Converts the NeMo MLlama model to HuggingFace format and saves it to the specified path.

        Args:
            output_path (Path): The path where the converted HuggingFace model will be saved.

        Returns:
            Path: The output path where the HuggingFace model was saved.
        """
        logging.info("Loading MLlama NeMo checkpoint. This may take a while...")
        source, source_config = self.ckpt_load(self)
        logging.info("MLlama NeMo checkpoint loaded.")
        logging.info("Initializing the HF model..")
        target = self.init()
        logging.info("Start Converting the model..")
        target = self.convert_state(source, target, source_config)
        target = target.cpu()
        target.save_pretrained(output_path)

        try:
            self.tokenizer.tokenizer.save_pretrained(output_path)
        except Exception:
            logging.warning("Failed to save tokenizer")

        print(f"Converted MLlama model saved to {output_path}")

        return output_path

    def convert_state(self, source, target, source_config):
        # pylint: disable=C0115,C0116,line-too-long
        """
        Maps and transforms the state dictionary from NeMo to HuggingFace format.

        Args:
            source: The source NeMo model.
            target: The target HuggingFace model.

        Returns:
            The target HuggingFace model with the converted state.
        """
        source = self._modify_mllama_source_state(source, source_config)
        mapping = {}
        transforms = []
        # Define the state mapping from NeMo to HuggingFace
        mapping.update(
            {
                "language_model.decoder.layers.*.self_attention.linear_proj.weight": "model.language_model.layers.*.self_attn.o_proj.weight",
                "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "model.language_model.layers.*.input_layernorm.weight",
                "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "model.language_model.layers.*.post_attention_layernorm.weight",
                "language_model.decoder.layers.*.mlp.linear_fc2.weight": "model.language_model.layers.*.mlp.down_proj.weight",
                "language_model.decoder.xattn_layers.*.cross_attention.q_layernorm.weight": "model.language_model.layers.*.cross_attn.q_norm.weight",
                "language_model.decoder.xattn_layers.*.cross_attention.linear_q.weight": "model.language_model.layers.*.cross_attn.q_proj.weight",
                "language_model.decoder.xattn_layers.*.cross_attention.k_layernorm.weight": "model.language_model.layers.*.cross_attn.k_norm.weight",
                "language_model.decoder.xattn_layers.*.cross_attention.linear_proj.weight": "model.language_model.layers.*.cross_attn.o_proj.weight",
                "language_model.decoder.final_layernorm.weight": "model.language_model.norm.weight",
                "language_model.output_layer.weight": "lm_head.weight",
            }
        )
        transforms.extend(
            [
                io.state_transform(
                    source_key="language_model.decoder.xattn_layers.*.gate_attn",
                    target_key="model.language_model.layers.*.cross_attn_attn_gate",
                    fn=_export_gate,
                ),
                io.state_transform(
                    source_key="language_model.decoder.xattn_layers.*.gate_ffn",
                    target_key="model.language_model.layers.*.cross_attn_mlp_gate",
                    fn=_export_gate,
                ),
                io.state_transform(
                    source_key="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
                    target_key=(
                        "model.language_model.layers.*.self_attn.q_proj.weight",
                        "model.language_model.layers.*.self_attn.k_proj.weight",
                        "model.language_model.layers.*.self_attn.v_proj.weight",
                    ),
                    fn=_export_text_qkv,
                ),
                io.state_transform(
                    source_key="language_model.decoder.layers.*.mlp.linear_fc1.weight",
                    target_key=(
                        "model.language_model.layers.*.mlp.gate_proj.weight",
                        "model.language_model.layers.*.mlp.up_proj.weight",
                    ),
                    fn=_export_simple_split,
                ),
                io.state_transform(
                    source_key="language_model.decoder.xattn_layers.*.cross_attention.linear_kv.weight",
                    target_key=(
                        "model.language_model.layers.*.cross_attn.k_proj.weight",
                        "model.language_model.layers.*.cross_attn.v_proj.weight",
                    ),
                    fn=_export_text_kv,
                ),
                io.state_transform(
                    source_key=(
                        "language_model.embedding.word_embeddings.weight",
                        "language_model.learnable_embedding.weight",
                    ),
                    target_key="model.language_model.embed_tokens.weight",
                    fn=_export_embedding_hf,
                ),
            ]
        )
        v = "vision_model.vision_encoder"
        mapping.update(
            {
                f"{v}.global_transformer.layers.*.self_attention.linear_proj.weight": "model.vision_model.global_transformer.layers.*.self_attn.o_proj.weight",
                f"{v}.global_transformer.layers.*.gate_attn": "model.vision_model.global_transformer.layers.*.gate_attn",
                f"{v}.global_transformer.layers.*.gate_ffn": "model.vision_model.global_transformer.layers.*.gate_ffn",
                f"{v}.global_transformer.layers.*.input_layernorm.bias": "model.vision_model.global_transformer.layers.*.input_layernorm.bias",
                f"{v}.global_transformer.layers.*.input_layernorm.weight": "model.vision_model.global_transformer.layers.*.input_layernorm.weight",
                f"{v}.global_transformer.layers.*.pre_mlp_layernorm.bias": "model.vision_model.global_transformer.layers.*.post_attention_layernorm.bias",
                f"{v}.global_transformer.layers.*.pre_mlp_layernorm.weight": "model.vision_model.global_transformer.layers.*.post_attention_layernorm.weight",
                f"{v}.global_transformer.layers.*.mlp.linear_fc1.bias": "model.vision_model.global_transformer.layers.*.mlp.fc1.bias",
                f"{v}.global_transformer.layers.*.mlp.linear_fc1.weight": "model.vision_model.global_transformer.layers.*.mlp.fc1.weight",
                f"{v}.global_transformer.layers.*.mlp.linear_fc2.bias": "model.vision_model.global_transformer.layers.*.mlp.fc2.bias",
                f"{v}.global_transformer.layers.*.mlp.linear_fc2.weight": "model.vision_model.global_transformer.layers.*.mlp.fc2.weight",
                f"{v}.transformer.layers.*.self_attention.linear_proj.weight": "model.vision_model.transformer.layers.*.self_attn.o_proj.weight",
                f"{v}.transformer.layers.*.input_layernorm.bias": "model.vision_model.transformer.layers.*.input_layernorm.bias",
                f"{v}.transformer.layers.*.input_layernorm.weight": "model.vision_model.transformer.layers.*.input_layernorm.weight",
                f"{v}.transformer.layers.*.pre_mlp_layernorm.bias": "model.vision_model.transformer.layers.*.post_attention_layernorm.bias",
                f"{v}.transformer.layers.*.pre_mlp_layernorm.weight": "model.vision_model.transformer.layers.*.post_attention_layernorm.weight",
                f"{v}.transformer.layers.*.mlp.linear_fc1.bias": "model.vision_model.transformer.layers.*.mlp.fc1.bias",
                f"{v}.transformer.layers.*.mlp.linear_fc1.weight": "model.vision_model.transformer.layers.*.mlp.fc1.weight",
                f"{v}.transformer.layers.*.mlp.linear_fc2.bias": "model.vision_model.transformer.layers.*.mlp.fc2.bias",
                f"{v}.transformer.layers.*.mlp.linear_fc2.weight": "model.vision_model.transformer.layers.*.mlp.fc2.weight",
                f"{v}.class_embedding": "model.vision_model.class_embedding",
                f"{v}.positional_embedding": "model.vision_model.gated_positional_embedding.embedding",
                f"{v}.gated_tile_positional_embedding.weight": "model.vision_model.gated_positional_embedding.tile_embedding.weight",
                f"{v}.gated_positional_embedding_gate": "model.vision_model.gated_positional_embedding.gate",
                f"{v}.ln_post.bias": "model.vision_model.layernorm_post.bias",
                f"{v}.ln_post.weight": "model.vision_model.layernorm_post.weight",
                f"{v}.ln_pre.bias": "model.vision_model.layernorm_pre.bias",
                f"{v}.ln_pre.weight": "model.vision_model.layernorm_pre.weight",
                f"{v}.post_tile_pos_embed.embedding.weight": "model.vision_model.post_tile_positional_embedding.embedding.weight",
                f"{v}.post_tile_pos_embed.gate": "model.vision_model.post_tile_positional_embedding.gate",
                f"{v}.pre_tile_pos_embed.embedding.weight": "model.vision_model.pre_tile_positional_embedding.embedding.weight",
                f"{v}.pre_tile_pos_embed.gate": "model.vision_model.pre_tile_positional_embedding.gate",
                "vision_model.vision_projection.encoder.bias": "model.multi_modal_projector.bias",
                "vision_model.vision_projection.encoder.weight": "model.multi_modal_projector.weight",
            }
        )
        transforms.extend(
            [
                io.state_transform(
                    source_key=(f"{v}.global_transformer.layers.*.self_attention.linear_qkv.weight"),
                    target_key=(
                        "model.vision_model.global_transformer.layers.*.self_attn.q_proj.weight",
                        "model.vision_model.global_transformer.layers.*.self_attn.k_proj.weight",
                        "model.vision_model.global_transformer.layers.*.self_attn.v_proj.weight",
                    ),
                    fn=_export_vision_qkv,
                ),
                io.state_transform(
                    source_key=(f"{v}.transformer.layers.*.self_attention.linear_qkv.weight"),
                    target_key=(
                        "model.vision_model.transformer.layers.*.self_attn.q_proj.weight",
                        "model.vision_model.transformer.layers.*.self_attn.k_proj.weight",
                        "model.vision_model.transformer.layers.*.self_attn.v_proj.weight",
                    ),
                    fn=_export_vision_qkv,
                ),
                io.state_transform(
                    source_key=f"{v}.conv1._linear.weight",
                    target_key="model.vision_model.patch_embedding.weight",
                    fn=_export_patch_embedding_hf,
                ),
            ]
        )
        return io.apply_transforms(source, target, mapping=mapping, transforms=transforms)

    @property
    def tokenizer(self) -> "TokenizerSpec":
        """
        Gets the tokenizer from the loaded model context.

        Returns:
            The tokenizer specification.
        """
        return io.load_context(str(self), subpath="model").tokenizer

    def ckpt_load(self, path: Path) -> Tuple[Dict, Dict]:
        """
        This function loads the state dict directly from a distributed checkpoint, and modify the state dict
        so that it is consistent with the key names you would get from loading the checkpoint into a model.
        This is a more memory-efficient method to obtain a state dict without initializing the nemo model.

        Args:
            path (Path): The path from which the model will be loaded.

        Returns
        -------
            Tuple[Dict, Dict]: The loaded state dict and the yaml config dict.
        """
        config = io.load_context(str(self), subpath="model.config")
        dist_ckpt_folder = path / "weights"
        state_dict = {}

        langauge_layers = config.language_model_config.num_layers
        vision_layers = config.vision_model_config.num_layers
        distributed_model_weights = load_distributed_model_weights(dist_ckpt_folder, True).items()
        for k, v in distributed_model_weights:
            if "_extra_state" in k:
                continue
            new_k = k.replace("module.", "")
            if "layers" in new_k and (v.size(0) == langauge_layers or v.size(0) == vision_layers):
                # Only split layers
                for i in range(v.size(0)):
                    state_dict[new_k.replace("layers", f"layers.{str(i)}")] = v[i]
            elif "global_transformer.layers" in new_k:
                for i in range(v.size(0)):
                    state_dict[new_k.replace("layers", f"layers.{str(i)}")] = v[i]
            state_dict[new_k] = v
        return state_dict, config

    def _modify_mllama_source_state(self, state_dict, source_config):
        """
        - Modify state dict to integrate cross-attention layers into self-attention layer.
        e.g. 11B: 32 self-attn + 8 cross-attn -> 40 layers, 90B: 80 self-attn + 20 cross-attn -> 100 layers
        - Change the layer index to match the cross_attention_layers in the model config.
        e.g. 11B: [3, 7, 11, 15, 19, 23, 27, 31] -> [3, 8, 13, 18, 23, 28, 33, 38]

        Args:
            state_dict: Source model state dict
            source_config: Model config dict

        Returns:
            _ModelState: Modified state
        """

        def convert_layer_num(match):
            layer_num = int(match.group(1))
            x_num = (layer_num - 3) // (cross_attention_frequency)
            if (layer_num - 3) % (cross_attention_frequency) == 0:
                new_layer_num = x_num + layer_num
                return f".{new_layer_num}."
            raise ValueError(
                f"Unexpected layer_num: {layer_num} (does not align with cross_attention_frequency={cross_attention_frequency})"
            )

        text_config = source_config.language_model_config
        cross_attention_frequency = text_config.num_layers // text_config.num_cross_attention_layers
        total_num_layer = text_config.num_layers + text_config.num_cross_attention_layers
        prefix = "language_model.decoder"

        new_state_dict = {}
        # Integrating layer indexes of self-attention and cross-attention
        for i in range(total_num_layer):
            cross_num = (i - 3) // (cross_attention_frequency + 1)
            if (i - 3) % (cross_attention_frequency + 1) == 0:
                xattn_index = cross_num * cross_attention_frequency + 3
                new_state_dict[f"{prefix}.layers.{i}.mlp.linear_fc1.layer_norm_weight"] = state_dict.pop(
                    f"{prefix}.xattn_layers.{xattn_index}.mlp.linear_fc1.layer_norm_weight"
                )
                new_state_dict[f"{prefix}.layers.{i}.mlp.linear_fc2.weight"] = state_dict.pop(
                    f"{prefix}.xattn_layers.{xattn_index}.mlp.linear_fc2.weight"
                )
                new_state_dict[f"{prefix}.layers.{i}.self_attention.linear_qkv.layer_norm_weight"] = state_dict.pop(
                    f"{prefix}.xattn_layers.{xattn_index}.cross_attention.linear_q.layer_norm_weight"
                )
                new_state_dict[f"{prefix}.layers.{i}.mlp.linear_fc1.weight"] = state_dict.pop(
                    f"{prefix}.xattn_layers.{xattn_index}.mlp.linear_fc1.weight"
                )
            else:
                attn_index = i - cross_num - 1
                new_state_dict[f"{prefix}.layers.{i}.mlp.linear_fc1.layer_norm_weight"] = state_dict.pop(
                    f"{prefix}.layers.{attn_index}.mlp.linear_fc1.layer_norm_weight"
                )
                new_state_dict[f"{prefix}.layers.{i}.mlp.linear_fc2.weight"] = state_dict.pop(
                    f"{prefix}.layers.{attn_index}.mlp.linear_fc2.weight"
                )
                new_state_dict[f"{prefix}.layers.{i}.self_attention.linear_qkv.layer_norm_weight"] = state_dict.pop(
                    f"{prefix}.layers.{attn_index}.self_attention.linear_qkv.layer_norm_weight"
                )
                new_state_dict[f"{prefix}.layers.{i}.mlp.linear_fc1.weight"] = state_dict.pop(
                    f"{prefix}.layers.{attn_index}.mlp.linear_fc1.weight"
                )

        for k, v in new_state_dict.items():
            state_dict[k] = v

        new_state_dict = {}
        # Align the cross-attention layer index with HF
        for k, v in state_dict.items():
            if "xattn_layers" in k:
                new_state_dict[re.sub(r"\.(\d+)\.", convert_layer_num, k)] = v
            else:
                new_state_dict[k] = v

        source = _ModelState(new_state_dict)
        return source

    @property
    def config(self) -> "HFMllamaConfig":
        """
        Generates the configuration for the HuggingFace MLlama model based on the NeMo model.

        Returns:
            HFMllamaConfig: A configuration object for the HuggingFace MLlama model.
        """
        source = io.load_context(str(self), subpath="model.config")
        vision_model_config = source.vision_model_config
        language_config = source.language_model_config

        vision_config = MllamaVisionConfig(
            num_hidden_layers=vision_model_config.num_layers,
            hidden_size=vision_model_config.hidden_size,
            attention_heads=vision_model_config.num_attention_heads,
            image_size=vision_model_config.vision_chunk_size,
            max_num_tiles=vision_model_config.vision_max_num_chunks,
            torch_dtype="bfloat16",
        )
        cross_attention_layers = [
            x + i
            for i, x in enumerate(language_config._init_fusion_schedule(language_config.num_cross_attention_layers))
        ]
        # Create text config for HuggingFace model
        text_config = MllamaTextConfig(
            rope_theta=language_config.rotary_base,
            num_hidden_layers=language_config.num_layers + language_config.num_cross_attention_layers,
            tie_word_embeddings=language_config.share_embeddings_and_output_weights,
            cross_attention_layers=cross_attention_layers,
            hidden_size=language_config.hidden_size,
            intermediate_size=language_config.ffn_hidden_size,
            num_attention_heads=language_config.num_attention_heads,
            num_key_value_heads=language_config.num_query_groups,
            vocab_size=language_config.vocab_size,
            rope_scaling={
                "factor": 8.0,
                "high_freq_factor": 4.0,
                "low_freq_factor": 1.0,
                "original_max_position_embeddings": 8192,
                "rope_type": "llama3",
            },
            eos_token_id=[128001, 128008, 128009],
            torch_dtype="bfloat16",
        )
        # Create the MllamaConfig for HuggingFace
        return HFMllamaConfig(vision_config=vision_config, text_config=text_config, torch_dtype="bfloat16")


def _rename_xattn_layer_nums_hf(source: Dict):
    def convert_layer_num(match):
        layer_num = int(match.group(1))
        cross_num = (layer_num - 3) // (cross_attention_frequency + 1)
        if (layer_num - 3) % (cross_attention_frequency + 1) == 0:
            new_layer_num = cross_num * cross_attention_frequency + 3
            return f"xattn_layers.{new_layer_num}."

        new_layer_num = layer_num - cross_num - 1
        return f"layers.{new_layer_num}."

    cross_attention_frequency = 4

    output_dict = {}
    for k, v in source.items():
        if "language_model" in k:
            output_dict[re.sub(r"layers\.(\d+)\.", convert_layer_num, k)] = v
        else:
            output_dict[k] = v
    return output_dict


def _import_embedding_hf(a):
    return torch.split(a, a.shape[0] - 8, dim=0)


def _import_patch_embedding_hf(a):
    return a.reshape(a.shape[0], -1)


def _import_gate(gate):
    return gate[0:1]


def _import_vision_qkv(ctx: io.TransformCTX, q, k, v):
    vision_config = ctx.target.config.vision_model_config

    head_num = vision_config.num_attention_heads
    num_query_groups = vision_config.num_query_groups
    head_size = vision_config.kv_channels
    hidden_size = vision_config.hidden_size
    return _merge_qkv(q, k, v, head_num, num_query_groups, head_size, hidden_size)


def _import_text_qkv(ctx: io.TransformCTX, q, k, v):
    text_config = ctx.target.config.language_model_config

    head_num = text_config.num_attention_heads
    num_query_groups = text_config.num_query_groups
    head_size = text_config.kv_channels
    hidden_size = text_config.hidden_size
    return _merge_qkv(q, k, v, head_num, num_query_groups, head_size, hidden_size)


def _import_text_kv(ctx: io.TransformCTX, k, v):
    text_config = ctx.target.config.language_model_config

    head_num = text_config.num_attention_heads
    num_query_groups = text_config.num_query_groups
    head_size = text_config.kv_channels
    hidden_size = text_config.hidden_size
    return _merge_kv(k, v, head_num, num_query_groups, head_size, hidden_size)


def _import_simple_concat(a, b):
    # for both (w1, w3) -> fc1, and (wk, wv) -> wkv
    return torch.cat((a, b), dim=0)


def _merge_kv(
    k: Tensor,
    v: Tensor,
    head_num: int,
    num_query_groups: int,
    head_size: int,
    hidden_size: int,
):
    old_tensor_shape = k.size()
    new_kv_tensor_shape = (num_query_groups, head_size) + old_tensor_shape[1:]

    k = k.view(*new_kv_tensor_shape)
    v = v.view(*new_kv_tensor_shape)

    kv_weights = torch.stack((k, v), dim=1)
    kv_weights = kv_weights.reshape(-1, *new_kv_tensor_shape[1:])
    assert kv_weights.ndim == 3, kv_weights.shape
    assert kv_weights.shape[0] == 2 * num_query_groups, kv_weights.shape
    assert kv_weights.shape[1] == head_size, kv_weights.shape
    assert kv_weights.shape[2] == old_tensor_shape[1], kv_weights.shape

    kv_weights = kv_weights.reshape([head_size * 2 * num_query_groups, hidden_size])
    return kv_weights


def _merge_qkv(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    head_num: int,
    num_query_groups: int,
    head_size: int,
    hidden_size: int,
):
    heads_per_group = head_num // num_query_groups
    old_tensor_shape = q.size()
    new_q_tensor_shape = (head_num, head_size) + old_tensor_shape[1:]
    new_kv_tensor_shape = (num_query_groups, head_size) + old_tensor_shape[1:]

    q = q.view(*new_q_tensor_shape)
    k = k.view(*new_kv_tensor_shape)
    v = v.view(*new_kv_tensor_shape)

    qkv_weights_l = []
    for i in range(num_query_groups):
        qkv_weights_l.append(q[i * heads_per_group : (i + 1) * heads_per_group, :, :])
        qkv_weights_l.append(k[i : i + 1, :, :])
        qkv_weights_l.append(v[i : i + 1, :, :])
    qkv_weights = torch.cat(qkv_weights_l)
    assert qkv_weights.ndim == 3, qkv_weights.shape
    assert qkv_weights.shape[0] == (heads_per_group + 2) * num_query_groups, qkv_weights.shape
    assert qkv_weights.shape[1] == head_size, qkv_weights.shape
    assert qkv_weights.shape[2] == old_tensor_shape[1], qkv_weights.shape

    qkv_weights = qkv_weights.reshape([head_size * (head_num + 2 * num_query_groups), hidden_size])

    return qkv_weights


def _split_kv(
    kv: Tensor,
    head_num: int,
    num_query_groups: int,
    head_size: int,
    hidden_size: int,
):
    kv_total_dim = 2 * num_query_groups

    linear_kv = kv.reshape([kv_total_dim, head_size, hidden_size])

    k_slice = torch.arange(0, kv_total_dim, 2)
    v_slice = torch.arange(1, kv_total_dim, 2)

    k_proj = linear_kv[k_slice].reshape(-1, hidden_size).cpu()
    v_proj = linear_kv[v_slice].reshape(-1, hidden_size).cpu()

    return k_proj, v_proj


def _split_qkv(qkv, head_num: int, num_query_groups: int, head_size: int, hidden_size: int):
    heads_per_group = head_num // num_query_groups
    qkv_total_dim = head_num + 2 * num_query_groups

    linear_qkv = qkv.reshape([qkv_total_dim, head_size, hidden_size])
    q_slice = torch.cat(
        [
            torch.arange((heads_per_group + 2) * i, (heads_per_group + 2) * i + heads_per_group)
            for i in range(num_query_groups)
        ]
    )
    k_slice = torch.arange(heads_per_group, qkv_total_dim, (heads_per_group + 2))
    v_slice = torch.arange(heads_per_group + 1, qkv_total_dim, (heads_per_group + 2))

    q_proj = linear_qkv[q_slice].reshape(-1, hidden_size).cpu()
    k_proj = linear_qkv[k_slice].reshape(-1, hidden_size).cpu()
    v_proj = linear_qkv[v_slice].reshape(-1, hidden_size).cpu()

    return q_proj, k_proj, v_proj


def _export_gate(gate):
    return gate[0:1]


def _export_patch_embedding_hf(a):
    return a.reshape(a.shape[0], 3, 14, 14)


def _export_vision_qkv(ctx: io.TransformCTX, qkv):
    vision_config = ctx.target.config.vision_config

    head_num = vision_config.attention_heads
    num_query_groups = vision_config.attention_heads
    hidden_size = vision_config.hidden_size
    head_size = hidden_size // head_num
    return _split_qkv(qkv, head_num, num_query_groups, head_size, hidden_size)


def _export_text_kv(ctx: io.TransformCTX, kv):
    text_config = ctx.target.config.text_config

    head_num = text_config.num_attention_heads
    num_query_groups = text_config.num_key_value_heads
    hidden_size = text_config.hidden_size
    head_size = hidden_size // head_num
    return _split_kv(kv, head_num, num_query_groups, head_size, hidden_size)


def _export_text_qkv(ctx: io.TransformCTX, qkv):
    text_config = ctx.target.config.text_config

    head_num = text_config.num_attention_heads
    num_query_groups = text_config.num_key_value_heads
    hidden_size = text_config.hidden_size
    head_size = hidden_size // head_num
    return _split_qkv(qkv, head_num, num_query_groups, head_size, hidden_size)


def _export_simple_split(linear_fc1):
    """Splits NeMo's fused MLP linear_fc1 weight into gate_proj and up_proj for HuggingFace format."""
    gate_proj, up_proj = torch.chunk(linear_fc1, 2, dim=0)
    return gate_proj, up_proj


def _export_embedding_hf(word_embeddings, learnable_embedding):
    """Transforms the word embeddings from NeMo to HuggingFace format."""
    return torch.cat((word_embeddings, learnable_embedding), dim=0)
